package com.atlassian.crucible.plugin.export;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpMethodBase;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.log4j.Logger;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import com.atlassian.crucible.spi.data.FisheyeReviewItemData;
import com.atlassian.crucible.spi.data.GeneralCommentData;
import com.atlassian.crucible.spi.data.ReviewData;
import com.atlassian.crucible.spi.data.VersionedLineCommentData;

/**
 * Provides helper methods to access the Crucible Web Services API.
 * 
 * @author Ross Rowe
 * 
 */
public class CrucibleRestClient implements CrucibleClient {

	private HttpClient client = new HttpClient();

	private static final String GET_REVIEW_URL = "/rest-service/reviews-v1/";

	private static final Logger logger = Logger
			.getLogger(CrucibleRestClient.class);

	private static final String LOGIN_URL = "/login";

	private String baseUrl;

	/**
	 * Performs a login action using the username and password.
	 */
	public String login(String userName, String password) {
		Map<String, String> parameterMap = new HashMap<String, String>();
		parameterMap.put("username", userName);
		parameterMap.put("password", password);
		return invokePost(LOGIN_URL, parameterMap);
	}

	/**
	 * Retrieves the {@link ReviewData} instance for the reviewId.
	 * 
	 * @param reviewId
	 * @return
	 */
	public ReviewData getReview(String reviewId) {
		String response = invokeGet(GET_REVIEW_URL + reviewId);
		try {
			JAXBContext context = JAXBContext.newInstance(ReviewData.class);
			Unmarshaller u = context.createUnmarshaller();
			JAXBElement<ReviewData> root = u.unmarshal(new StreamSource(
					new ByteArrayInputStream(response.getBytes())),
					ReviewData.class);
			return root.getValue();
		} catch (JAXBException e) {
			logger.error(e);
		}
		return null;
	}

	/**
	 * Invokes a HTTP POST method and returns the response contents.
	 * 
	 * @param url
	 * @param parameterMap
	 * @return
	 */
	private String invokePost(String url, Map<String, String> parameterMap) {
		Set<String> keys = parameterMap.keySet();
		PostMethod method = new PostMethod(baseUrl + url);
		for (String key : keys) {
			method.addParameter(key, parameterMap.get(key));
		}
		try {
			return invoke(method);
		} catch (HttpException e) {
			logger.error(e);
		} catch (IOException e) {
			logger.error(e);
		}
		return null;
	}

	/**
	 * Invokes a HTTP GET method and returns the response contents.
	 * 
	 * @param url
	 * @return
	 */
	private String invokeGet(String url) {
		try {
			GetMethod method = new GetMethod(baseUrl + url);
			return invoke(method);
		} catch (MalformedURLException e) {
			logger.error(e);
		} catch (IOException e) {
			logger.error(e);
		}
		return null;
	}

	/**
	 * Invokes a HttpMethod and returns the response contents.
	 * 
	 * @param method
	 * @return
	 * @throws IOException
	 * @throws HttpException
	 */
	private String invoke(HttpMethodBase method) throws IOException,
			HttpException {
		String contents = null;
		int statusCode = client.executeMethod(method);
		if (statusCode != -1) {
			contents = method.getResponseBodyAsString();
			method.releaseConnection();
		}
		return contents;
	}

	/**
	 * Returns the list of {@link GeneralCommentData} instances for the
	 * reviewId.
	 * 
	 * @param token
	 * @param reviewId
	 * @return
	 */
	public List<GeneralCommentData> getGeneralComments(String reviewId) {
		String response = invokeGet(GET_REVIEW_URL + reviewId + "/comments");
		List<GeneralCommentData> comments = new ArrayList<GeneralCommentData>();
		try {
			XPath xpath = XPathFactory.newInstance().newXPath();
			InputSource inputSource = new InputSource(new ByteArrayInputStream(
					response.getBytes()));
			NodeList nodes = (NodeList) xpath.evaluate("//generalComment",
					inputSource, XPathConstants.NODESET);
			for (int i = 0; i < nodes.getLength(); i++) {
				Node node = nodes.item(i);
				JAXBContext context = JAXBContext
						.newInstance(GeneralCommentData.class);
				Unmarshaller u = context.createUnmarshaller();
				JAXBElement<GeneralCommentData> root = u.unmarshal(node,
						GeneralCommentData.class);
				comments.add(root.getValue());
			}
		} catch (XPathExpressionException e) {
			logger.error(e);
		} catch (JAXBException e) {
			logger.error(e);
		}
		return comments;
	}

	/**
	 * Returns a list of {@link VersionedLineCommentData} instances for the
	 * reviewId.
	 * 
	 * @param token
	 * @param reviewId
	 * @return
	 */
	public List<VersionedLineCommentData> getVersionedComments(String reviewId) {
		String response = invokeGet(GET_REVIEW_URL + reviewId + "/comments");
		List<VersionedLineCommentData> comments = new ArrayList<VersionedLineCommentData>();
		try {
			XPath xpath = XPathFactory.newInstance().newXPath();
			InputSource inputSource = new InputSource(new ByteArrayInputStream(
					response.getBytes()));
			NodeList nodes = (NodeList) xpath.evaluate("//versionedComment",
					inputSource, XPathConstants.NODESET);
			for (int i = 0; i < nodes.getLength(); i++) {
				Node node = nodes.item(i);
				JAXBContext context = JAXBContext
						.newInstance(VersionedLineCommentData.class);
				Unmarshaller u = context.createUnmarshaller();
				JAXBElement<VersionedLineCommentData> root = u.unmarshal(node,
						VersionedLineCommentData.class);
				comments.add(root.getValue());
			}
		} catch (XPathExpressionException e) {
			logger.error(e);
		} catch (JAXBException e) {
			logger.error(e);
		}
		return comments;
	}

	/**
	 * 
	 * @param baseUrl
	 */
	public void setBaseUrl(String baseUrl) {
		this.baseUrl = baseUrl;
	}

	/**
	 * Retrieves the {@link FisheyeReviewItemData} instance for the reviewId and
	 * the reviewItemId
	 * @param reviewId
	 * @param reviewItemId
	 * @return
	 */
	public FisheyeReviewItemData getReviewItemData(String reviewId,
			String reviewItemId) {
		String response = invokeGet(GET_REVIEW_URL + reviewId + "/reviewitems/"
				+ reviewItemId);
		try {
			JAXBContext context = JAXBContext
					.newInstance(FisheyeReviewItemData.class);
			Unmarshaller u = context.createUnmarshaller();
			JAXBElement<FisheyeReviewItemData> root = u.unmarshal(
					new StreamSource(new ByteArrayInputStream(response
							.getBytes())), FisheyeReviewItemData.class);
			return root.getValue();
		} catch (JAXBException e) {
			logger.error(e);
		}

		return null;
	}

	/**
	 * Returns the userId (there isn't a mechanism within the Crucible REST API
	 * to retrieve the user name).
	 */
	public String getUserName(String userId) {
		return userId;
	}

}
